unit MJCore;
{$I MJBase.inc}

// ****************************************************************************
// **  SPIELFELDGENERATOREN                                                  **
// ****************************************************************************

procedure InitGameValues;  // Spielfeld resetten
var b:byte;
begin
  // Phase 1: Tile-Basisdaten anlegen
  for b:=1 to 34 do begin TilesLeft[b]:=4;
    TileFitMin[b]:=b;TileFitMax[b]:=b;end;
  for b:=35 to 38 do begin TilesLeft[b]:=1;
    TileFitMin[b]:=35;TileFitMax[b]:=38;end;
  for b:=39 to 42 do begin TilesLeft[b]:=1;
    TileFitMin[b]:=39;TileFitMax[b]:=42;end;

  // Phase 2: diverse Variablen neu setzen
  MaxTiles:=0;MaxPlanes:=0;MaxTileX:=0;MaxTileY:=0;
  Marked:=False;Stuck:=False;
  HistPos:=0;HistMax:=0;
  Tiles[0].Active:=True;
  Time_NextRefresh:=0;Time_ActiveSince:=Timer;Time_Offset:=0;
end;

// +++++++++++++++ GenerateGame_Classic: klassischer (dummer) Spielfeldgenerator

procedure GenerateGame_Classic;
var x,y:integer;

procedure InitTile(x,y:integer;Plane:byte);
  { sucht ein passendes Sprite und legt ein Tile damit an }
var Take:byte;
begin
  repeat Take:=random(TileSprites)+1;
  until TilesLeft[Take]>0;
  dec(TilesLeft[Take]);
  AddTile(x,y,Plane,Take);
end;

begin CleverlyGenerated:=false;
  // Spielfeld zeilenweise aufbauen
  for x:=12 downto 1 do InitTile(x*TileWidth,0,1);
  for x:=10 downto 3 do InitTile(x*TileWidth,TileHeight,1);
  for x:=11 downto 2 do InitTile(x*TileWidth,2*TileHeight,1);
  for x:=12 downto 1 do InitTile(x*TileWidth,3*TileHeight,1);
  InitTile(14*TileWidth,3*TileHeight+TileHeight SHR 1,1);
  InitTile(13*TileWidth,3*TileHeight+TileHeight SHR 1,1);
  for x:=12 downto 1 do InitTile(x*Tilewidth,4*TileHeight,1);
  InitTile(0,3*TileHeight+TileHeight SHR 1,1);
  for x:=11 downto 2 do InitTile(x*TileWidth,5*TileHeight,1);
  for x:=10 downto 3 do InitTile(x*TileWidth,6*TileHeight,1);
  for x:=12 downto 1 do InitTile(x*TileWidth,7*TileHeight,1);

  for y:=1 to 6 do for x:=9 downto 4 do InitTile(x*TileWidth,y*TileHeight,2);
  for y:=2 to 5 do for x:=8 downto 5 do InitTile(x*TileWidth,y*TileHeight,3);
  for y:=3 to 4 do for x:=7 downto 6 do InitTile(x*TileWidth,y*TileHeight,4);
  InitTile(6*TileWidth+TileWidth SHR 1,3*TileHeight+TileHeight SHR 1,5);
end;

// +++++ GenerateGame_Clever: neuer Algorithmus, der unlsbare Spiele verhindert

procedure GenerateGame_Clever; 
const MaxTrials=10;
var x,y:integer;
    Trial:integer;

function FindFreePosition:integer;
  // sucht eine freie Position zum Anlegen eines Steines
const MaxTrials=100;
var Trial:integer;
    TileNo:integer;
    Accepted:boolean;
    CTile:TTile; NCheck:integer;
    UnderlyingSlots,UnderlyingTiles:integer;
    OnThisLine:integer;
begin
  Trial:=0;
  repeat inc(Trial);
    repeat TileNo:=random(MaxTiles)+1;
    until not(Tiles[TileNo].Active);
    CTile:=Tiles[TileNo];

    // Zwingende Bedingung: unter dem neuen Stein MUSS bereits ein Stein liegen
    Accepted:=(CTile.Plane=1);
    if not Accepted then begin
      UnderlyingSlots:=0; UnderlyingTiles:=0;
      for NCheck:=1 to MaxTiles do
        if NCheck<>TileNo then with Tiles[NCheck] do
          if (Plane<CTile.Plane) AND
             (CTile.X<X+TileWidth) AND (CTile.Y<Y+TileHeight) AND
             (CTile.X>X-TileWidth) AND (CTile.Y>Y-TileHeight) then begin
            inc(UnderlyingSlots);
            if Active then inc(UnderlyingTiles);
          end;
      Accepted:=(UnderlyingSlots=UnderlyingTiles);
    end;

    // Seitenbedingungen: Entweder links oder rechts mu auch ein Stein liegen
    if Accepted then begin
      Accepted:=false; OnThisLine:=0;
      NCheck:=0;repeat inc(NCheck);
        if NCheck<>TileNo then with Tiles[NCheck] do
          if Active AND (CTile.Plane=Plane) AND
             (CTile.Y<Y+TileHeight) AND (CTile.Y>Y-TileHeight) then begin
            inc(OnThisLine);
            if Abs(CTile.X-X)<=TileWidth then Accepted:=true;
          end;
      until Accepted OR (NCheck>=MaxTiles);
      if OnThisLine=0 then Accepted:=true;
    end;

  until (Trial>MaxTrials) OR Accepted;
  if Accepted then FindFreePosition:=TileNo else FindFreePosition:=0;
end;

function TryToFill:boolean;  // versucht, das Spielfeld mit Steinen aufzufllen
var Sprite1,Sprite2:integer;
    Tile1,Tile2:integer;
begin
  for Tile1:=1 to MaxTiles do Tiles[Tile1].Active:=false;
  TryToFill:=false;
  TotalTiles:=0;
  repeat
    Tile1:=FindFreePosition;
    if Tile1=0 then exit;
    repeat Sprite1:=random(TileSprites)+1;
    until TilesLeft[Sprite1]>0;
    dec(TilesLeft[Sprite1]);
    Tiles[Tile1].Sprite:=Sprite1;
    Tiles[Tile1].Active:=true;

    Tile2:=FindFreePosition;
    if Tile2=0 then exit;
    repeat Sprite2:=random(TileFitMax[Sprite1]-TileFitMin[Sprite1]+1)+TileFitMin[Sprite1];
    until TilesLeft[Sprite2]>0;
    dec(TilesLeft[Sprite2]);
    Tiles[Tile2].Sprite:=Sprite2;
    Tiles[Tile2].Active:=true;

    inc(TotalTiles,2);
  until TotalTiles>=MaxTiles;
  TryToFill:=true;
end;

// GenerateGame_Clever Hauptroutine
begin CleverlyGenerated:=true;
  // "leeres" Spielfeld zeilenweise aufbauen
  for x:=12 downto 1 do AddTile(x*TileWidth,0,1,0);
  for x:=10 downto 3 do AddTile(x*TileWidth,TileHeight,1,0);
  for x:=11 downto 2 do AddTile(x*TileWidth,2*TileHeight,1,0);
  for x:=12 downto 1 do AddTile(x*TileWidth,3*TileHeight,1,0);
  AddTile(14*TileWidth,3*TileHeight+TileHeight SHR 1,1,0);
  AddTile(13*TileWidth,3*TileHeight+TileHeight SHR 1,1,0);
  for x:=12 downto 1 do AddTile(x*Tilewidth,4*TileHeight,1,0);
  AddTile(0,3*TileHeight+TileHeight SHR 1,1,0);
  for x:=11 downto 2 do AddTile(x*TileWidth,5*TileHeight,1,0);
  for x:=10 downto 3 do AddTile(x*TileWidth,6*TileHeight,1,0);
  for x:=12 downto 1 do AddTile(x*TileWidth,7*TileHeight,1,0);

  for y:=1 to 6 do for x:=9 downto 4 do AddTile(x*TileWidth,y*TileHeight,2,0);
  for y:=2 to 5 do for x:=8 downto 5 do AddTile(x*TileWidth,y*TileHeight,3,0);
  for y:=3 to 4 do for x:=7 downto 6 do AddTile(x*TileWidth,y*TileHeight,4,0);
  AddTile(6*TileWidth+TileWidth SHR 1,3*TileHeight+TileHeight SHR 1,5,0);

  // Spielfeld auffllen; ggf. Spielfeld klassisch erzeugen
  Trial:=0;
  repeat inc(Trial);
  until (Trial>=MaxTrials) OR TryToFill;
  if Trial>=MaxTrials then begin
    InitGameValues;
    GenerateGame_Classic;
  end;
end;

// ++++++++++++++++++++++++++++++++++++++++++++++++++++++ NewGame: Hauptroutine

procedure NewGame;  { Neues Spiel beginnen }
begin
  InitGameValues;
  if AvoidUnsolveableGames then GenerateGame_Clever
                           else GenerateGame_Classic;
  TotalTiles:=MaxTiles;Changed:=True;Stuck:=False;Won:=False;Game:=True;
end;

// ****************************************************************************
// **  SONSTIGE MAHJONGG-SPEZIFISCHE ROUTINEN                                **
// ****************************************************************************

procedure MoveCheck;  { prft Tiles auf Bewegbarkeit, zhlt mgliche Zge }
var ntile,ncheck:integer;result:boolean;ctile:ttile;
    nsprite,csprite:byte;AlreadyWon:boolean;
begin
  AlreadyWon:=Won;
  maxactive:=0;
  for ntile:=1 to maxtiles do if tiles[ntile].active then begin
    ctile:=tiles[ntile];

    { prft, ob kein anderes Tile auf CTile=Tiles[NTile] liegt }
    result:=true;ncheck:=0;repeat inc(ncheck);
      if ncheck<>ntile then with tiles[ncheck] do
        if active AND (ctile.plane<plane) AND
           (ctile.x<x+tilewidth) AND (ctile.x>x-tilewidth) AND
           (ctile.y<y+tilewidth) AND (ctile.y>y-tilewidth)
           then result:=false;
    until not(result) OR (ncheck>=maxtiles);

    if result then begin
      { prft, ob kein Tile links von CTile liegt }
      result:=true;ncheck:=0;repeat inc(ncheck);
        if ncheck<>ntile then with tiles[ncheck] do
          if active AND (ctile.plane=plane) AND
             (ctile.x>x+tilewidth-1) AND
             (ctile.y<y+tilewidth) AND (ctile.y>y-tilewidth)
             then result:=false;
      until not(result) OR (ncheck>=maxtiles);

      { prft, ob kein Tile rechts von CTile liegt }
      if not result then begin
        result:=true;ncheck:=0;repeat inc(ncheck);
          if ncheck<>ntile then with tiles[ncheck] do
            if active AND (ctile.plane=plane) AND
               (ctile.x<x-tilewidth+1) AND
               (ctile.y<y+tilewidth) AND (ctile.y>y-tilewidth)
               then result:=false;
        until not(result) OR (ncheck>=maxtiles);
      end;
    end;

    tiles[ntile].moveable:=result;
    if result then begin
      inc(maxactive);active[maxactive]:=ntile;
    end;
  end;

  { sucht und zhlt mgliche Zge }
  moves:=0;
  for ntile:=1 to maxactive do begin
    nsprite:=tiles[active[ntile]].sprite;
    for ncheck:=1 to maxactive do if ntile<>ncheck then begin
      csprite:=tiles[active[ncheck]].sprite;
      if (nsprite>=tilefitmin[csprite]) AND
         (nsprite<=tilefitmax[csprite]) then
         inc(moves);
    end;
  end;
  Moves:=Moves SHR 1;
  Stuck:=(Moves=0);
  Won:=Stuck AND (MaxTiles=0);
  WonBuildMode:=True;
  if Won AND not(AlreadyWon) then ScoreForm.Execute(Time_Delta);
end;

procedure UseTile(what:word);  { Mausklick auf ein Tile }
begin
  if not Marked then begin
    if not Tiles[what].Moveable then exit;
    Marked:=True;MarkTile:=what;
    Enlightenment(TilePos(what));
  exit;end;
  if (what=MarkTile )OR not(Tiles[what].Moveable) then begin
    Marked:=False;
  exit;end;
  if (Tiles[what].Sprite<TileFitMin[Tiles[MarkTile].Sprite]) OR
     (Tiles[what].Sprite>TileFitMax[Tiles[MarkTile].Sprite]) then begin
    MarkTile:=what;
  exit;end;
  Marked:=False;
  KillTile(MarkTile);KillTile(what);
  inc(HistPos);HistMax:=HistPos;
  History[HistMax,0]:=Time_Delta;
  History[HistMax,1]:=MarkTile;
  History[HistMax,2]:=what;
  Changed:=True;
end;

procedure Hint;
var NTile,NCheck:integer;NSprite,CSprite:byte;OK:boolean;
begin
  if Stuck then exit;

  if GaveHint then begin
    KillTile(HintTile1);
    KillTile(HintTile2);
    inc(HistPos);HistMax:=HistPos;
    History[HistMax,0]:=Time_Delta;
    History[HistMax,1]:=HintTile1;
    History[HistMax,2]:=HintTile2;
    MoveCheck;Redraw;
  exit;end;

  OK:=False;
  NTile:=MaxActive+1;repeat dec(NTile);
    NSprite:=Tiles[Active[NTile]].Sprite;
    NCheck:=MaxActive+1;repeat dec(NCheck);
      if NTile<>NCheck then begin
        CSprite:=Tiles[Active[Ncheck]].Sprite;
        if (NSprite>=TileFitMin[CSprite]) AND
           (NSprite<=TileFitMax[CSprite]) then
           OK:=True;
      end;
    until (NCheck<=1) OR OK;
  until OK OR (NTile<=1);
  HintTile1:=Active[NTile];
  HintTile2:=Active[NCheck];
  GaveHint:=True;
  Enlightenment(TilePos(HintTile1));
  Enlightenment(TilePos(HintTile2));
  Time_Offset:=Time_Delta+HintPenalty;Time_ActiveSince:=Timer;
  Flip;
end;

end.


